"""
This module provides API like service for the anime site chiaanime.tv
It used selenium for the final step in the process of getting information
if anime is to be downloaded, as the download link is rendered by
javascript.
This currently only supports Firefox headless versions.
Use of selenium creates an added overhead and may slow down the script
and thus it is recommended that you have atleast a moderate 
internet connection.

"""



import os 

from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
import requests 
from bs4 import BeautifulSoup




class ChiaAnime():
    """API To chiaanime.tv site"""

    def __init__(self):
        self._base_url = 'http://ww2.chia-anime.tv'
        self.session = requests.Session()
   
  
    def search(self,anime):
        """
        Searches the anime 
        :param anime: anime name to be search
        :return: list of dictionaries, each consisting 'name' and 'url' as key or None. 
        """
        s = self.session 
        url = '{base_url}/search/{search_item}'.format(base_url=self._base_url,search_item=anime)

        res = s.get(url)
        res.raise_for_status()
        soup = BeautifulSoup(res.text,'lxml')   # add lxml to requirements.
        raw_result = soup.select('[class~=title] > a')       # Results.
        
        # if raw_result is empty, list will be empty, thus will evaluate to None
        result = [
                    {
                        'name':elem.text,
                        'url':elem.get('href',None),
                    }
                    for elem in raw_result
                ]
        return result


    def show_genres(self):
        """
        List all the genres available in the website

        :return: List of dictionaries with, 'name' and 'url'.
        """
        s = self.session
        url = '{base_url}/index'.format(base_url=self._base_url)

        res = s.get(url)
        res.raise_for_status()
        soup = BeautifulSoup(res.text,'lxml')
        #result in raw form (includes the tags also in the 
        # formed list).
        raw_result = soup.select('p[style] > a')
        
        
        result = [
                    {
                        'name':elem.text, #extracting the genre name
                        'url' :elem.get('href',None) #getting the url (None if not found)
                    }
                    for elem in raw_result
                ]
        return result
            

    def under_genre(self,genre_name):
        """
        Extract information under specific genres

        :param genre_name: name of the genre
        :return: list of dictionaries, with 'name' and 'url' of anime or None
        """
        s = self.session
        genre_name = genre_name.casefold() # case insensitive
        add_url = '?genre={genre_name}'.format(genre_name=genre_name)
        url = '{base_url}/{add_url}'.format(base_url=self.base_url,add_url=add_url)
        res = s.get(url)
        soup = BeautifulSoup(res.text,'lxml')
        
        raw_result = soup.select('h3 > a[style]')
        result = [
                    {
                        'name':elem.text,
                        'url' :elem.get('href',None)
                    }
                    for elem in raw_result
        ]
        return result


    def most_popular(self):
        """
        Lists the most-popular animes on the website

        :return: List of dictionaries, with 'name' and 'url'.
        """
        s = self.session
        popular = 'most-popular-anime-series/'
        url = '{base_url}/{added_info}'.format(base_url=self._base_url,added_info=popular)

        res = s.get(url)  
        res.raise_for_status()  # raises error with connectivity
        soup = BeautifulSoup(res.text,'lxml')
        raw_result = soup.select('div[style] h3 > a')

        result = [
                    {'name':elem.text,
                     'url' :elem.get('href',None),
                    }
                    for elem in raw_result
        ]
        return result


    def anime_info(self,url=None):
        """
        Provides the info available for an anime.

        :param url: Url to the anime info needed
        :return: dictionary of *'info', 'value', where info is the information
        """
        s = self.session
        res = s.get(url)
        res.raise_for_status()
        soup = BeautifulSoup(res.text,'lxml')


        # Due to ineffecient site building, and ineffecient coding, following is done :/
        anime_info = soup.select('blockquote > div')
        anime_plot = soup.select('blockquote > p')

        paired_values = [i.text.strip().split(':') for i in anime_info]
        paired_values.append(['plot',anime_plot[0].text.strip()])

        # The above give something like [['english','some value'],['another','another_value']
        
       
        result_dict = {
                    i[0].casefold():' '.join(i[1:]).strip() for i in paired_values
        }
        # The above converts the paired values to dict.

        return result_dict


    def get_episodes_link(self,anime_url,start,end):
        """
        Generator function yielding episodes_download_url

        :param anime_url: This is the main page url
        :param range: A string of number specifying the range
                      of episodes.
                      Default range is 1, only the first episode url
        :yield: episode_download_url
        """

        s = self.session
        res = s.get(anime_url)
        soup = BeautifulSoup(res.text,'lxml')
        all_episode_links = soup.select('h3[itemprop] > a')
        
        # Episode links are in reverse order
        all_episode_links.reverse()

        
        driver = self._get_browser()
        for i in all_episode_links[start-1: end]:
            '''
            This for loop do not check if the end range is valid.
            For end range beyond the links, all links are included uptill last link.
            '''
            episode_url = i.get('href')
            episode_download_urls = self._extract_download_url(episode_url,driver)
            yield episode_download_urls
        driver.quit()
    
    ########################################
    # Following functions are used only for# 
    # implementation details.              #
    #######################################

    def _url_to_anime_premium(self,episode_url):
        """Get the anime premium site links
        :param episode_url: (str) url of the episode
        :return: Download url of the episode 
        """
        s = self.session
        res = s.get(episode_url)
        soup = BeautifulSoup(res.text,'lxml')
        # Finding anime_premium download site link
        raw_result = soup.find(id='download')
        url = raw_result.get('href')
        return url


    # This is used to add support for multiple browsers.
    @staticmethod
    def _get_browser(browser='firefox'):
        """
        Internal usage, gets the browser to be used for scraping
        :return: A selenium webdriver
        """
        
        if browser.casefold() == 'firefox':
            driver_options = webdriver.FirefoxOptions()
            driver_options.add_argument('--headless')
            driver = webdriver.Firefox(options=driver_options)
            return driver


    def _extract_download_url(self,episode_url,browser):
        """Gets the download url for the episode
        :param episode_url: url (str) of the episode
        :param return: (str) of download url
        """
        
        anime_premium_link = self._url_to_anime_premium(episode_url)
        driver = browser
        driver.get(anime_premium_link)
        # There must be exception catching here 
        elements = WebDriverWait(driver, 30).until(
        EC.presence_of_all_elements_located((By.CSS_SELECTOR, "a[target='_blank']"))
                            )
        name = os.path.basename(episode_url.strip('/'))
        # This gathers all the download links for episode
        # e.g. Server 0, server 1 etc.
        episode_download_urls = {
                                name : [elem.get_attribute('href') for elem in elements]
        }
        
        return episode_download_urls
    

    @property
    def base_url(self):
        return self._base_url


    @base_url.setter
    def base_url(self,url):
        #make it raise an exception if invalid url given
        self._base_url = url
    


"""
Downloader interface
Class is used to ease the process of adding features such as
archiving the file, and using database, multithreaded downloading
asynch download.

"""
import os
import re
import requests
from bs4 import BeautifulSoup

from tqdm import tqdm

class Downloader():
    '''
    Sequentially downloads.
    '''

    def __init__(self,folder):
        self.session = requests.Session()
        self.download_folder_path = folder
    
    
    def run(self,url):
        '''
        Main function that downloads the url.
        '''
        session = self.session
        file_name = list(url.keys())[0]
        file_path = '{folder}/{episode_name}'.format(
                                folder = self.download_folder_path,
                                episode_name=file_name
        )
        current_ep_url = url[file_name][0] # First server is used
        print("Episode: {}".format(file_name))
        try:
            downloaded_file_size = os.path.getsize(file_path)
        except FileNotFoundError:
            downloaded_file_size = 0
        headers = {
                "Range":"bytes={}-".format(downloaded_file_size)
        }
        with open(file_path,'ab') as f:
            with session.get(current_ep_url,stream=True,headers=headers) as s:
                current_file_size = int(s.headers['Content-Length'])
                total_file_size = current_file_size + downloaded_file_size 
                print('\nTotal size of file: {}B'.format(total_file_size))
                bars = int(total_file_size)//1024
                pos = downloaded_file_size//1024
                for chunk in tqdm( 
                                iterable = s.iter_content(chunk_size=1024),
                                initial=pos,
                                total = bars,
                                unit = 'KB',
                                unit_scale=True,
                                unit_divisor=1024,
                                 ):
                    if chunk:
                        f.write(chunk)










                



"""
Code in this module is very ugly, look at your own risk.


Maybe use the paginated display.

"""
import sys


def _output(result):
    """User interface, Needs changing """
    print('='*34)
    print('WELCOME, Your search results are : ')
    print('='*34)
    for i,data in enumerate(result):
        num = i+1
        print('{index}. {data_name}'.format(index=num,data_name=data.get('name')))
    

    return 

def _selecting_anime():
    """One function interface to select anime"""
    # This s nippet is hangable(infinite loop), if user continously enters blank.
    while True:
        move_ahead = input("\nDownload or access anime info, [y/n]? ")
        
        if not move_ahead:  # nothing entered
            continue

        if move_ahead.casefold() == 'y':
            # It might be good to use some library, maybe textwrap ? o.O
            print('\n')
            print('+'*34)
            print('Available options are: ')
            print('+'*34)

            print('\n[anime number] ([i] | [d]) These are positional argument.')
            print('i: Show anime info')
            print('d: Download the season')
            print('q: Quit')
            print('\ne.g. 1 i (info of 1st anime) or 1 d (download) \n')
            
            available_options = {
                                'd',
                                'i'
                                }
            # check for EOFError: not added right now.
            while True:
                option = input('Enter your option: ')
                option_list = option.strip().split(' ',1)
                if option.casefold()=='q':
                    sys.exit('Bye')
                if len(option_list) == 2:
                    if not option_list[1] in available_options:
                        print("No argument {} available, please try again".format(option_list[1]))
                        continue
                    num,option = (  (int(option_list[0])), 
                                    (option_list[1])
                                )
                    break
                else:
                    print('{option} is wrong, please try again'.format(option=option))
                    continue
            break

        if move_ahead.casefold() == 'n':
            sys.exit('Okay will see you later.')
            
    return num,option
    
def _anime_extractor(s_result : list):
    '''
    It does something.
    '''
    _output(s_result)

    return _selecting_anime()
                
def _genre_extractor(g_result : list):
    _output(g_result)

    while True:
        move_ahead = input("Hey, want to see animes under specific genre? [y/n] (n for quit): ")
        if move_ahead.casefold() == 'n':
            sys.exit('Exiting')

        if move_ahead.casefold() == 'y':
            genre = input("Enter the genre\n")
            break
        else:
            continue

    return genre.casefold()

def extract(passed_result,extractor=None):
    """
    Extracts and outputs the text from the given list
    e in the e_result stands for extractor result ;)
    """


    # You have to ensure only one of the extractors are given,
    #  or as the arguments specify.
    if extractor=='anime':
        return _anime_extractor(passed_result)

    if extractor =='genres':
        return _genre_extractor(passed_result)
    

def _selecting_info(info : dict):
    """
    Used to select anime from the user, to show information
    """
    move_ahead = '' 
    while True:
        console_string = """Enter the info name: [[q] quit; [b] back; [d] download]\n\n"""
        option = input(console_string).strip()

        if not option:
            continue
        if option.casefold()=='q':
            sys.exit("Bye for now ;)")
        if option.casefold() not in info:
            if option.casefold() == 'd':
                move_ahead = 'download'
                break
            elif option.casefold() == 'b':
                move_ahead = 'back'
                break
            print("'{}' wrong info\n".format(option))
            continue

        value = info[option.casefold()]
        if not value:
            print("Sorry but information for {} is  not updated by site".format(option))
        print(value)
        continue
    return move_ahead

 
def anime_info_output(info : dict):
    #change is required, output not in good formatting
        print("\nWelcome to the info section of anime")
        print("Available info:")
        for i,val in enumerate(info.keys()):
            num = i + 1
            print("\t{index}. {val}".format(index=num,val=val))
        # sys will exit in _selecting_info if no info is selected
        return _selecting_info(info)

def download_info(episodes=None):
    """Interface to Download
    
    :return: a tuple of integers
    """
    if not episodes:
        # Indirect download invoked.
        episodes = input("Enter the episode numbers\ne.g. 45-47 will download episodes from 45 to 47(including)\n[q to quit]\n")
        if episodes.casefold() == 'q':
            sys.exit('Bye :D')
        
    # Stripping of white spaces
    stripped_episodes = episodes.strip()
    tupled_episodes = tuple(map(int,stripped_episodes.split('-')))
    #Get the *integer* indexes from stripped string
    try:
        ep_start,ep_end = tupled_episodes
    except ValueError: # if only one episode entered
        ep_start = tupled_episodes[0] # Only get the first episode.
        ep_end = ep_start
    return ep_start, ep_end





"""
The Driver program
Currently the only site supported is chiaanime.tv

""" 
import sys
import os

 
from .api import ChiaAnime 
from .args import get_args 
from .interface import (extract, 
                        anime_info_output,
                        download_info)

from .downloader import Downloader
 
class Anime():
    ''' 
    Main Class
    '''
    def __init__(self):
        self.chiaanime = ChiaAnime()
        self.args = get_args()
        # used to store anime search result.
        self._result = None

    def _anime_processing(self,anime_result,anime_num=None,option=None):
        """
        Common interface for processing anime
        result.
        """

        if not anime_num and not option:
            anime_num, option = extract(anime_result,extractor='anime')
        url = anime_result[anime_num-1].get('url')
        if option=='i':
            self.anime_info(url)
        
        if option=='d':
            self.anime_download(url)


    def _searcher(self,anime_name,index=None,choice=None):
        """
        A common interface for downloading,info and searching
        This is used only for implementation details.

        :param anime_name: A str of anime name to be searched
        :param index: Search index to be used.
        :param choice: Used to direct the flow of program from either
                       info or download.
        """
        anime_result = self.chiaanime.search(anime_name)
        # Temporary storing anime_result for further processing.
        self.result = anime_result

        if not anime_result:
            sys.exit("Sorry,no results found for {anime_name}, maybe try its japanese name ".format(anime_name=anime_name))
        self._anime_processing(anime_result,index,choice)

    def _create_download_folder(self):
        """
        A single channel to create a folder.
        Provides easy interface for scalable folder operations
        """
        folder = self.args.folder
        cur_dir = os.getcwd()
        folder_path = os.path.join(cur_dir,folder)
        os.makedirs(folder_path,exist_ok=True)
        return folder_path


    def anime_download(self,anime_url): 
        """Download anime
        :param anime_url: Url to animes main page
        
        """
        folder = self._create_download_folder()
        print("+"*34)
        print("Downloaded folder is: {}".format(folder))
        print("To change rerun the program with --folder option")
        print('+'*34)

        if self.args.download_anime:
        # Download is directly invoked, hence use the --range
        # argument.
            ep_range = download_info(self.args.episode_range)

        else:
        # Download is indirectly invoked, give 
        # user an interface for specifying range
            ep_range = download_info()
        anime_name = os.path.basename(anime_url.strip('/'))
        print('\nDownloading {}, please wait'.format(anime_name))

        #extracting episode ranges
        download_url_generator = self.chiaanime.get_episodes_link
        downloader = Downloader(folder)
        
        for url in download_url_generator(anime_url,*ep_range):
            downloader.run(url)
        sys.exit('Bye :D')

    def anime_info(self,anime_url):
        '''
        Provides the information about an anime, 
        :param anime_url: main url of the anime
        ''' 
        result_info = self.chiaanime.anime_info(url=anime_url)
        option = anime_info_output(result_info)    
        if option == 'download':
            self.anime_download(anime_url)

        if option == 'back':
            # move back to the displayed list of animes
            # Thus, imitating the _searcher, but instead of again getting the
            # search result, use the temporary search result
            self._anime_processing(self.result)
        sys.exit('Bye :D')

    def run(self):
        """
        Main function
        """
        args = self.args
        
        if args.search: # add a waiting message, probably use the threading.(could be fun :D)
            anime = args.search
            self._searcher(anime)

        if args.download_anime:
            anime = args.download_anime
            index = args.search_index
            self._searcher(anime,index,choice='d')
        
        if args.info:
            anime = args.info
            index = args.search_index
            self._searcher(anime,index,choice='i')
        

        if args.genres: 
            genres_result = self.chiaanime.show_genres()
            genre_name = extract(genres_result,extractor='genres')
            under_genre_result = self.chiaanime.under_genre(genre_name=genre_name)
            self._anime_processing(under_genre_result)

 
        if args.most_popular:
            anime_result = self.chiaanime.most_popular()
            self._anime_processing(anime_result)
        
        
        
    @property
    def result(self):
        return self._result
    
    @result.setter
    def result(self,val):
        self._result = val
        
def main():
    """
    Main console_scipt entrpy point 
    """
    anime = Anime()
    anime.run()